/*
 * Startup.S : processes after CPU RESET, from WinARM example
 *   by uratan! 2008.9.8
 */

/*
 * Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs
 */
	.set Mode_USR, 0x10
	.set Mode_FIQ, 0x11
	.set Mode_IRQ, 0x12
	.set Mode_SVC, 0x13
	.set Mode_ABT, 0x17
	.set Mode_UND, 0x1B
	.set Mode_SYS, 0x1F

	.set I_Bit, 0x80    /* when I bit is set, IRQ is disabled */
	.set F_Bit, 0x40    /* when F bit is set, FIQ is disabled */

/*
 * Pre-defined interrupt handlers that may be directly
 * overwritten by C interrupt functions
 */
	.extern	FIQ_handler
	.extern	UNDEF_handler
	.extern	PABT_handler
	.extern	DABT_handler


/*===================================================================*/

/*
 * Stack Configuration (Stack Sizes in Bytes)
 */
	.set UND_Stack_Size, 0x00000080		/* Stack for Undefined Mode */
	.set ABT_Stack_Size, 0x00000080		/* Stack for Abort Mode */
	.set FIQ_Stack_Size, 0x00000080		/* Stack for Fast Interrupt Mode */
	.set IRQ_Stack_Size, 0x00000200		/* Stack for Interrupt Mode */
	.set SVC_Stack_Size, 0x00000080		/* Stack for Supervisor Mode */
	.set USR_Stack_Size, 0x00000800		/* Stack for User/System Mode */

.arm
.section .stack, "w"
.align 4
	.space	(USR_Stack_Size+3)&~3
	.space	(SVC_Stack_Size+3)&~3
	.space	(IRQ_Stack_Size+3)&~3
	.space	(FIQ_Stack_Size+3)&~3
	.space	(ABT_Stack_Size+3)&~3
	.space	(UND_Stack_Size+3)&~3
Top_Stack:


/*===================================================================*/

.text

/*
 * Exception Vectors, Mapped to Address 0.
 * Absolute addressing mode must be used.
 */
.arm
.section .vectorg, "ax"
__Vectors:
	LDR		PC, Reset_Addr
	LDR		PC, Undef_Addr
	LDR		PC, SWI_Addr
	LDR		PC, PAbt_Addr
	LDR		PC, DAbt_Addr
	NOP							/* Reserved Vector (holds ISP checksum) */
	LDR		PC, [PC, #-0x0FF0]		/* Vector from VicVectAddr register */
	LDR		PC, FIQ_Addr

Reset_Addr:	.word	Reset_Handler		/* in myself */
Undef_Addr:	.word	UNDEF_handler		/* in main.c */
SWI_Addr:	.word	SoftwareInterrupt	/* in swi_handler.S */
PAbt_Addr:	.word	PABT_handler		/* in main.c */
DAbt_Addr:	.word	DABT_handler		/* in main.c */
IRQ_Addr:	.word	UNDEF_handler		/* (use VIC auto-vectoring) */
FIQ_Addr:	.word	FIQ_handler			/* in main.c */
			.word	0		/* rounds the vectors and ISR addresses to 64 bytes total */

.size   __Vectors, . - __Vectors


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * system startup processes
 */

.arm
.section .init, "ax"
.global __startup
.func __startup
__startup:

Reset_Handler:
	/*
	 * Setup Stack for each mode
	 */
	LDR		R0, =Top_Stack

	MSR		CPSR_c, #Mode_UND|I_Bit|F_Bit
	MOV		SP, R0
	SUB		R0, R0, #UND_Stack_Size

	MSR		CPSR_c, #Mode_ABT|I_Bit|F_Bit
	MOV		SP, R0
	SUB		R0, R0, #ABT_Stack_Size

	MSR		CPSR_c, #Mode_FIQ|I_Bit|F_Bit
	MOV		SP, R0
	SUB		R0, R0, #FIQ_Stack_Size

	MSR		CPSR_c, #Mode_IRQ|I_Bit|F_Bit
	MOV		SP, R0
	SUB		R0, R0, #IRQ_Stack_Size

	MSR		CPSR_c, #Mode_SVC|I_Bit|F_Bit
	MOV		SP, R0
	SUB		R0, R0, #SVC_Stack_Size

//	MSR		CPSR_c, #Mode_USR				/* Interrupts enabled */
//	MSR		CPSR_c, #Mode_USR|I_Bit|F_Bit	/* Interrupts disabled */
//	MSR		CPSR_c, #Mode_SYS				/* Interrupts enabled */
	MSR		CPSR_c, #Mode_SYS|I_Bit|F_Bit	/* Interrupts disabled */
	MOV		SP, R0

	/*
	 * Relocate .data section (Copy from ROM to RAM)
	 */
	LDR		R1, =_etext
	LDR		R2, =_data
	LDR		R3, =_edata
	CMP		R2, R3
	BEQ		DataIsEmpty
LoopRel:
	CMP		R2, R3
	LDRLO	R0, [R1], #4
	STRLO	R0, [R2], #4
	BLO		LoopRel
DataIsEmpty:

	/*
	 * Clear .bss section (Zero init)
	 */
	MOV		R0, #0
	LDR		R1, =__bss_start__
	LDR		R2, =__bss_end__
	CMP		R1,R2
	BEQ		BSSIsEmpty
LoopZI:
	CMP		R1, R2
	STRLO	R0, [R1], #4
	BLO		LoopZI
BSSIsEmpty:

	/*
	 * Enter the C code
	 */
	B		main

.size   __startup, . - __startup
.endfunc


.end
